__version__ = "0.0.2"

from ._core import ArrayAxesStandardizer, CorrectDrift, ROIRect
from ._sample_data import sample_2d, sample_3d, sample_3d_ch
from ._widget import CorrectDriftDock, TableWidget

__all__ = (
    "sample_2d",
    "sample_3d",
    "sample_3d_ch",
    "CorrectDriftDock",
    "TableWidget",
    "CorrectDrift",
    "ArrayAxesStandardizer",
    "ROIRect",
)



import numpy as np

from napari_correct_drift import CorrectDriftDock, TableWidget


# make_napari_viewer is a pytest fixture that returns a napari viewer object
# capsys is a pytest fixture that captures stdout and stderr output streams
def test_correct_drift_widget(make_napari_viewer, capsys):
    # make viewer and add an image layer using our fixture
    viewer = make_napari_viewer()
    viewer.add_image(np.random.random((100, 100)))

    # create our widget, passing in the viewer
    my_widget = CorrectDriftDock(viewer)

    assert 1 == 1


def test_table_widget():
    random_drifts = np.random.rand(16, 3)
    # viewer = make_napari_viewer()

    tw = TableWidget(viewer=None)
    tw.set_content(random_drifts)
    tw.save(filename="tmp.csv")

    tw.load(filename="tmp.csv")

    assert np.allclose(random_drifts, tw.get_content())



import numpy as np

from napari_correct_drift._core import CorrectDrift, ROIRect
from napari_correct_drift._sample_data import Example_TYX, Example_TZYX


def test_tyx():
    example = Example_TYX()
    img_2d = example.create()
    ist_2d = CorrectDrift(img_2d, "tyx")

    for t0 in [0, 17]:
        ofs = ist_2d._estimate_drift_absolute(t0)
        ok = ofs - example.shifts + example.shifts[t0]
        assert np.allclose(ok, 0), "Shift estimation mismatch. Diff:\n"


def test_tzyx():
    example = Example_TZYX()
    img_3d = example.create()
    ist_3d = CorrectDrift(img_3d, "tzyx")

    for t0 in [0, 16]:
        ofs = ist_3d._estimate_drift_absolute(t0)
        ok = ofs - example.shifts + example.shifts[t0]
        assert np.allclose(ok, 0), "Shift estimation mismatch. Diff:\n"


def test_tcyx():
    example = Example_TYX()
    img_2d = example.create()
    img_2d = np.stack([img_2d, img_2d // 2], axis=1)
    ist_2d = CorrectDrift(img_2d, "tcyx")

    for t0 in [0, 7]:
        ofs = ist_2d._estimate_drift_absolute(t0, channel=0)
        ok = ofs - example.shifts + example.shifts[t0]
        assert np.allclose(ok, 0), "Shift estimation mismatch. Diff:\n"


def test_tczyx():
    example = Example_TZYX()
    img_3d = example.create()

    img_3d = np.stack([img_3d, img_3d // 2], axis=1)
    ist_3d = CorrectDrift(img_3d, "tczyx")

    for t0 in [0, 16]:
        ofs = ist_3d._estimate_drift_absolute(t0, channel=0)
        ok = ofs - example.shifts + example.shifts[t0]
        assert np.allclose(ok, 0), "Shift estimation mismatch."


def test_tyx_roi():
    example = Example_TYX()
    img_2d = example.create()
    ist_2d = CorrectDrift(img_2d, "tyx")

    poly2d = np.array(
        [
            [23.0, 122.57642757, 142.27071085],
            [23.0, 122.57642757, 190.77701484],
            [23.0, 181.65929409, 190.77701484],
            [23.0, 181.65929409, 142.27071085],
        ]
    )

    roi = ROIRect.from_shape_poly(poly2d, "tyx", 0, 1)
    t0 = roi.t0

    ofs = ist_2d._estimate_drift_absolute(t0, roi=roi)
    ok = ofs - example.shifts + example.shifts[t0]
    assert np.allclose(ok, 0), "Shift estimation mismatch."


def test_tzyx_roi():
    example = Example_TZYX()
    img_3d = example.create()
    ist_3d = CorrectDrift(img_3d, "tzyx")

    poly_3d = np.array(
        [
            [11.0, 7.0, 59.46602839, 35.94993883],
            [11.0, 7.0, 59.46602839, 66.41732531],
            [11.0, 7.0, 89.637615, 66.41732531],
            [11.0, 7.0, 89.637615, 35.94993883],
        ]
    )

    roi = ROIRect.from_shape_poly(poly_3d, "tzyx", 7, 15)
    t0 = roi.t0

    ofs = ist_3d._estimate_drift_absolute(t0, roi=roi)
    ok = ofs - example.shifts + example.shifts[t0]
    assert np.allclose(ok, 0), "Shift estimation mismatch."


def test_tczyx_roi():
    example = Example_TZYX()
    img_3d = example.create()

    img_3d = np.stack([img_3d, img_3d[::-1]], axis=1)

    ist_3d = CorrectDrift(img_3d, "tczyx")

    poly_3d = np.array(
        [
            [11.0, 0, 2.0, 59.46602839, 35.94993883],
            [11.0, 0, 7.0, 59.46602839, 66.41732531],
            [11.0, 0, 7.0, 89.637615, 66.41732531],
            [11.0, 0, 11.0, 89.637615, 35.94993883],
        ]
    )

    roi = ROIRect.from_shape_poly(poly_3d, "tzyx", 7, 15)
    t0 = roi.t0
    c0 = roi.c0
    ofs = ist_3d._estimate_drift_absolute(t0, roi=roi)

    ofs = ist_3d._estimate_drift_absolute(t0, roi=roi, channel=c0)
    ok = ofs - example.shifts + example.shifts[t0]
    assert np.allclose(ok, 0)


def test_tyx_rel():
    example = Example_TYX()
    img_2d = example.create()
    ist_2d = CorrectDrift(img_2d, "tyx")

    for t0 in [0, 6, 13]:
        ofs = ist_2d._estimate_drift_relative(t0)
        ok = ofs - example.shifts + example.shifts[t0]
        assert np.allclose(ok, 0)


def test_tczyx_rel():
    example = Example_TZYX()
    img_3d = example.create()

    img_3d = np.stack([img_3d, img_3d[:, ::-1] // 2], axis=1)

    ist_3d = CorrectDrift(img_3d, "tczyx")

    for t0 in [0, 16]:
        ofs = ist_3d._estimate_drift_relative(t0, channel=0)
        ok = ofs - example.shifts + example.shifts[t0]
        assert np.allclose(ok, 0), str(ok)


def test_tyx_roi_rel():
    example = Example_TYX()
    img_2d = example.create()
    ist_2d = CorrectDrift(img_2d, "tyx")

    poly2d = np.array(
        [
            [23.0, 122.57642757, 142.27071085],
            [23.0, 122.57642757, 190.77701484],
            [23.0, 181.65929409, 190.77701484],
            [23.0, 181.65929409, 142.27071085],
        ]
    )

    start_roi = ROIRect.from_shape_poly(poly2d, "tyx", 0, 1)
    t0 = start_roi.t0

    ofs = ist_2d._estimate_drift_relative(t0, roi=start_roi)
    ok = ofs - example.shifts + example.shifts[t0]
    assert np.allclose(ok, 0), str(ok)


def test_zcyx_roi_rel():
    example = Example_TZYX()
    img_3d = example.create()

    img_3d = np.stack([img_3d, img_3d[::-1]], axis=1)
    ist_3d = CorrectDrift(img_3d, "tczyx")

    poly_3d = np.array(
        [
            [11.0, 0, 2.0, 59.46602839, 35.94993883],
            [11.0, 0, 7.0, 59.46602839, 66.41732531],
            [11.0, 0, 7.0, 89.637615, 66.41732531],
            [11.0, 0, 11.0, 89.637615, 35.94993883],
        ]
    )

    roi = ROIRect.from_shape_poly(poly_3d, "tzyx", 7, 15)
    t0 = roi.t0
    c0 = roi.c0
    assert t0 == 11
    assert c0 == 0
    assert roi.z_min == 7
    assert roi.z_max == 15

    ofs = ist_3d._estimate_drift_relative(t0, roi=roi)

    ofs = ist_3d._estimate_drift_relative(t0, roi=roi, channel=c0)
    ok = ofs - example.shifts + example.shifts[t0]

    assert np.allclose(ok, 0), str(ok)


def test_tyx_big_roi_rel():
    example = Example_TYX()
    img_2d = example.create()
    ist_2d = CorrectDrift(img_2d, "tyx")

    poly2d = np.array(
        [
            [23.0, 2.57642757, 4.27071085],
            [23.0, 2.57642757, 250.77701484],
            [23.0, 250.65929409, 250.77701484],
            [23.0, 250.65929409, 4.27071085],
        ]
    )

    start_roi = ROIRect.from_shape_poly(poly2d, "tyx", 0, 1)
    t0 = start_roi.t0

    ofs = ist_2d._estimate_drift_relative(t0, roi=start_roi)

    ok = ofs - example.shifts + example.shifts[t0]
    assert np.allclose(ok, 0)


def test_roirect():
    poly_3d = np.array(
        [
            [11.0, 3.0, 11.0, 59.46602839, 35.94993883],
            [11.0, 3.0, 11.0, 59.46602839, 66.41732531],
            [11.0, 3.0, 11.0, 89.637615, 66.41732531],
            [11.0, 3.0, 11.0, 89.637615, 35.94993883],
        ]
    )

    roi = ROIRect.from_shape_poly(poly_3d, "tczyx", 7, 15)
    t0 = roi.t0
    c0 = roi.c0
    assert t0 == 11
    assert c0 == 3
    assert roi.z_min == 7
    assert roi.z_max == 15
    assert roi.y_max == 90
    assert roi.x_min == 36

    roi_clone = ROIRect.from_bbox(roi.bbox, t0, c0)
    assert np.allclose(roi.bbox, roi_clone.bbox)
    assert np.allclose(roi.c0, roi_clone.c0)



"""
This module is an example of a barebones sample data provider for napari.

It implements the "sample data" specification.
see: https://napari.org/stable/plugins/guides.html?#sample-data

Replace code below according to your needs.
"""
from __future__ import annotations

import numpy as np
from skimage.draw import disk
from skimage.morphology import ball, binary_dilation


class Example_TZYX:
    def __init__(self):
        deltas_3d = (
            [[1, 6, 4]] * 4
            + [[-1, 4, 1]] * 6
            + [[1, 2, -3]] * 6
            + [[0, -2, 7]] * 4
            + [[0, 2, -5]] * 4
            + [[0, 7, 0]] * 4
            + [[0, 0, -7]] * 3
        )

        shifts_3d = np.cumsum(deltas_3d, 0)
        shifts_3d = np.vstack(
            [
                [0] * 3,
                shifts_3d,
            ]
        )

        self.shape = (32, 16, 128, 80)
        self.start_pos = (8, 24, 32)
        self.deltas = deltas_3d
        self.shifts = shifts_3d

    def create(self, noise_level=0):
        deltas = np.asarray(self.deltas)

        T, Z, Y, X = self.shape
        img = np.zeros((T, Z, Y, X), dtype="bool")

        t = 0
        img[t][self.start_pos] = 1
        img[t] = binary_dilation(img[t], ball(3))

        pos = np.array(self.start_pos, dtype="int32")

        for d in deltas:
            t += 1
            pos += d
            img[t][tuple(pos)] = 1
            img[t] = binary_dilation(img[t], ball(3))

        img = img.astype("uint8") * (255 - noise_level)

        for t in range(T):
            img[t] += (np.random.rand(Y, X) * noise_level).astype("uint8")

        return img


class Example_TYX:
    def __init__(self):
        deltas_2d = [[8, 8]] * 10
        deltas_2d += [[-2, 6]] * 5
        deltas_2d += [[6, 3]] * 10
        deltas_2d += [[0, 7]] * 3
        deltas_2d += [[11, -3]] * 3

        deltas_2d = [(0, y, x) for y, x in deltas_2d]

        shifts_2d = np.cumsum(deltas_2d, 0)
        shifts_2d = np.vstack([[0, 0, 0], shifts_2d])

        self.shape = (32, 256, 256)
        self.start_pos = (32, 32)
        self.deltas = deltas_2d
        self.shifts = shifts_2d

    def create(self, noise_level=0):
        deltas = np.asarray(self.deltas)

        T, Y, X = self.shape
        img = np.zeros((T, Y, X), dtype="uint8")

        for t in range(T):
            img[t] += (np.random.rand(Y, X) * noise_level).astype("uint8")

        t = 0
        rr, cc = disk(self.start_pos, 8, shape=(256, 256))
        img[t, rr, cc] = 255

        pos = np.array(self.start_pos)

        for d in deltas:
            t += 1
            pos += d[1:]
            rr, cc = disk(pos, 8, shape=(256, 256))
            img[t, rr, cc] = 255

        return img


class Example_TCYX(Example_TYX):
    def create(self, noise_level=0):
        img = super().create(noise_level=noise_level)

        img = np.stack([img, img[::-1], np.zeros_like(img)], axis=1)
        self.shape = img.shape

        return img


def sample_2d():
    return [
        (
            Example_TYX().create(noise_level=16),
            {"name": "test_TYX", "contrast_limits": (0, 64)},
        ),
    ]


def sample_3d():
    return [
        (
            Example_TZYX().create(noise_level=16),
            {"name": "test_TZYX", "contrast_limits": (0, 64)},
        ),
    ]


def sample_3d_ch():
    return [
        (
            Example_TCYX().create(noise_level=16),
            {
                "name": "test_TCYX",
                "contrast_limits": (0, 64),
            },
        ),
    ]



"""
This module is an example of a barebones QWidget plugin for napari

It implements the Widget specification.
see: https://napari.org/stable/plugins/guides.html?#widgets

Replace code below according to your needs.
"""
from typing import TYPE_CHECKING

import numpy as np
from napari.layers.image.image import Image as IMAGE_LAYER
from napari.layers.shapes.shapes import Shapes as SHAPE_LAYER
from napari.utils import notifications
from pandas import DataFrame, read_csv
from qtpy.QtCore import QAbstractTableModel, Qt
from qtpy.QtWidgets import (
    QCheckBox,
    QComboBox,
    QFileDialog,
    QGridLayout,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QSpinBox,
    QTableView,
    QVBoxLayout,
    QWidget,
)

from ._core import CorrectDrift, ROIRect

if TYPE_CHECKING:
    import napari


class CorrectDriftDock(QWidget):
    ROI_LAYER_NAME = "ROI"

    def _init_input_layer_selection(self):
        ### Input Image layer selection
        self.input_layer = QComboBox()
        self.input_image_layers = []

        def init_input_layer():
            self.input_layer.clear()
            self.input_image_layers = []
            for layer in self.viewer.layers:
                if isinstance(layer, IMAGE_LAYER):
                    self.input_image_layers.append(layer)
                    self.input_layer.addItem(layer.name)

        init_input_layer()

        def on_inserted_layer(event):
            layer = event.value

            if isinstance(layer, IMAGE_LAYER):
                self.input_image_layers.append(layer)
                self.input_layer.addItem(layer.name)

        self.viewer.layers.events.inserted.connect(on_inserted_layer)

        def on_removed_layer(event):
            if event.value == self.get_current_input_layer():
                init_input_layer()

        self.viewer.layers.events.removed.connect(on_removed_layer)

        ### Axes selection

        axis_panel = QWidget()
        axis_layout = QGridLayout()
        axis_panel.setLayout(axis_layout)

        self.axis_combos = {}

        def update_axes_selection(event=None):
            layer = self.get_current_input_layer()

            *extra_dims, _, _ = layer.data.shape

            self.axis_combos = {}
            for d, d_size in enumerate(extra_dims):
                tmp = QComboBox()
                tmp.addItems(
                    [
                        f"Time ({d_size})",
                        f"Channels ({d_size})",
                        f"Z-axis ({d_size})",
                    ]
                )
                tmp.setCurrentIndex(d)
                self.axis_combos[d] = tmp

            for i in reversed(range(axis_layout.count())):
                axis_layout.itemAt(i).widget().setParent(None)

            for d, combo in self.axis_combos.items():
                axis_layout.addWidget(QLabel(f"Axis {d} is: "), d + 1, 0)
                axis_layout.addWidget(combo, d + 1, 1)
                combo.setToolTip(
                    f"Define which dimension Axis {d} corresponds to. 'Time' needs to be selected"
                )

        self.input_layer.currentIndexChanged.connect(update_axes_selection)

        if len(self.viewer.layers) > 0:
            update_axes_selection(self.viewer.layers[0])

        ### add both to dock
        input_panel = QGroupBox("Input & Axes")
        input_layout = QVBoxLayout()
        input_panel.setLayout(input_layout)

        input_layout.addWidget(self.input_layer)
        input_layout.addWidget(axis_panel)

        self.main_layout.addWidget(input_panel)

    def _init_run_panel(self):
        ### add both to dock
        run_panel = QGroupBox("Correct Drift")
        run_layout = QVBoxLayout()
        run_panel.setLayout(run_layout)

        tmp_panel = QWidget()
        tmp_layout = QHBoxLayout()
        tmp_panel.setLayout(tmp_layout)

        # estimate button
        self.estimate_drift_button = QPushButton("Estimate Drift")
        tmp_layout.addWidget(self.estimate_drift_button)
        self.estimate_drift_button.clicked.connect(self.estimate_drift)

        # tmp_layout.addWidget(QLabel("  or "))

        # load button
        self.load_drift_button = QPushButton("Load Drift")
        tmp_layout.addWidget(self.load_drift_button)
        self.load_drift_button.clicked.connect(self.load_drift)

        run_layout.addWidget(tmp_panel)

        # apply
        self.apply_drift_button = QPushButton("Apply Drift")
        tmp_layout.addWidget(self.apply_drift_button)
        self.apply_drift_button.clicked.connect(self.apply_drift)

        self.main_layout.addWidget(run_panel)

    def get_current_input_layer(self):
        return self.input_image_layers[self.input_layer.currentIndex()]

    def _init_key_and_roi_panel(self):
        # tracking panel
        key_and_roi_panel = QGroupBox("Key frames")
        key_and_roi_layout = QGridLayout()
        key_and_roi_panel.setLayout(key_and_roi_layout)
        i = 1

        # stabilize relative to
        self.estimate_drift_mode = QComboBox()
        self.estimate_drift_mode.addItem("previous frame")
        self.estimate_drift_mode.addItem("absolute frame")

        key_and_roi_layout.addWidget(QLabel("Relative to: "), i, 0)
        key_and_roi_layout.addWidget(self.estimate_drift_mode, i, 1)
        i += 1

        # key frame
        self.key_frame = QSpinBox()
        self.key_frame.setMinimum(0)
        self.key_frame.setValue(0)

        key_and_roi_layout.addWidget(QLabel("Key frame"), i, 0)
        key_and_roi_layout.addWidget(self.key_frame, i, 1)
        i += 1

        # key channel
        self.key_channel = QSpinBox()
        self.key_channel.setMinimum(0)
        self.key_channel.setValue(0)

        key_and_roi_layout.addWidget(QLabel("Key channel"), i, 0)
        key_and_roi_layout.addWidget(self.key_channel, i, 1)
        i += 1

        # use roi
        self.roi_checkbox = QCheckBox()
        self.roi_checkbox.setChecked(False)

        def add_shapes_layer(checked):
            if checked:
                for layer in self.viewer.layers:
                    if isinstance(layer, SHAPE_LAYER) and (
                        layer.name == self.ROI_LAYER_NAME
                    ):
                        break
                else:
                    ndim = self.get_current_input_layer().ndim
                    self.viewer.add_shapes(name=self.ROI_LAYER_NAME, ndim=ndim)
                    notifications.show_info(
                        f"Add a ROI to the created '{self.ROI_LAYER_NAME}' shape layer. \nThe ROI's key frame / channel will be used"
                    )

        def toggle_z_selection(checked):
            for w in self.z_roi_sel_widgets:
                w.setVisible(checked)

            self.key_frame.setDisabled(checked)
            self.key_channel.setDisabled(checked)

        self.roi_checkbox.toggled.connect(add_shapes_layer)
        self.roi_checkbox.toggled.connect(toggle_z_selection)

        key_and_roi_layout.addWidget(QLabel("Use ROI: "), i, 0)
        key_and_roi_layout.addWidget(self.roi_checkbox, i, 1)
        i += 1

        # ROI z-min and z=max
        self.roi_z_min = QSpinBox()
        self.roi_z_min.setMinimum(0)
        self.roi_z_min.setValue(0)

        self.roi_z_max = QSpinBox()
        self.roi_z_max.setMinimum(0)
        self.roi_z_max.setValue(0)

        self.roi_z_min_label = QLabel("Roi Z-min")
        key_and_roi_layout.addWidget(self.roi_z_min_label, i, 0)
        key_and_roi_layout.addWidget(self.roi_z_min, i, 1)
        i += 1

        self.roi_z_max_label = QLabel("Roi Z-max")
        key_and_roi_layout.addWidget(self.roi_z_max_label, i, 0)
        key_and_roi_layout.addWidget(self.roi_z_max, i, 1)
        i += 1

        # used later to toggle visibility
        self.z_roi_sel_widgets = [
            self.roi_z_min_label,
            self.roi_z_min,
            self.roi_z_max_label,
            self.roi_z_max,
        ]

        for w in self.z_roi_sel_widgets:
            w.setVisible(False)

        self.main_layout.addWidget(key_and_roi_panel)

    def __init__(self, napari_viewer: "napari.Viewer"):
        self.viewer = napari_viewer
        super().__init__()

        # main class for drift correction
        self.CorrDrift = None

        self.main_layout = QVBoxLayout()

        self._init_input_layer_selection()

        self._init_run_panel()

        self._init_key_and_roi_panel()

        self._init_other_params()

        self._init_tool_tips()

        self.main_layout.setAlignment(Qt.AlignTop)

        self.setLayout(self.main_layout)

    def _init_tool_tips(self):
        self.input_layer.setToolTip("Select input layer for drift correction")

        self.estimate_drift_button.setToolTip("Estimate drift")
        self.load_drift_button.setToolTip("Load drift from .csv")
        self.apply_drift_button.setToolTip("Apply drift shown in dirft table")

        self.estimate_drift_mode.setToolTip(
            "Mode for drift estimation:\n"
            " - relative: estimate from previous frame. ROI will move along!\n"
            " - absolute: estimate against absolute frame                    "
        )

        self.key_frame.setToolTip(
            "The frame number that should be stabilized during drift correction.\n"
            "When ROI is enabled, the value is inferred from the ROI.            "
        )
        self.key_channel.setToolTip(
            "The channel number that should be stabilized during drift correction.\n"
            "When ROI is enabled, the value is inferred from the ROI.              "
        )

        self.roi_checkbox.setToolTip(
            "Use ROI from 'ROI' shape layer. By default 1st ROI is used"
        )
        self.roi_z_min.setToolTip("Minimum z-plane for the ROI")
        self.roi_z_max.setToolTip("Maximum z-plane for the ROI")

        self.increment_box.setToolTip(
            "Time increment for drift estimation. Useful for faster estimation and \n"
            "slow drifts. Skipped frames will be linearly interpolated.              "
        )
        self.upsample_box.setToolTip(
            "Subpixel drift estimation. Useful for slow drifts"
        )
        self.extend_output.setToolTip(
            "Apply drifts with extended spatial dimensions. The raw image frames will\n"
            "be fully contained in the output.                                         "
        )

    def _init_other_params(self):
        parameter_panel = QGroupBox("Parameters")
        parameter_layout = QGridLayout()
        parameter_panel.setLayout(parameter_layout)
        i = 1

        # increment
        self.increment_box = QSpinBox()
        self.increment_box.setMinimum(1)
        self.increment_box.setValue(1)

        parameter_layout.addWidget(QLabel("Increment"), i, 0)
        parameter_layout.addWidget(self.increment_box, i, 1)
        i += 1

        # subpixel
        self.upsample_box = QSpinBox()
        self.upsample_box.setMinimum(1)
        self.upsample_box.setMaximum(20)
        self.upsample_box.setValue(1)

        parameter_layout.addWidget(QLabel("Upsample factor"), i, 0)
        parameter_layout.addWidget(self.upsample_box, i, 1)
        i += 1

        # subpixel
        self.extend_output = QCheckBox()
        self.extend_output.setChecked(False)

        parameter_layout.addWidget(QLabel("Extend output"), i, 0)
        parameter_layout.addWidget(self.extend_output, i, 1)
        i += 1

        parameter_panel.setLayout(parameter_layout)
        self.main_layout.addWidget(parameter_panel)

    def _check_input(self):
        layer = self.get_current_input_layer()
        image = layer.data

        # get axis
        dims = ""
        for _, combo in self.axis_combos.items():
            dims += "tcz"[combo.currentIndex()]
        dims += "yx"

        if "t" not in dims:
            raise RuntimeError("Drift correction axis 'Time' not set...")

        if len(set(dims)) != len(dims):
            raise RuntimeError(
                "Dimensions are not unique. Choose each dimensions only once"
            )

        self.CorrDrift = CorrectDrift(
            image,
            dims,
        )

        return self.CorrDrift

    def estimate_drift(self):
        self._check_input()

        # fixed frame
        key_frame = self.key_frame.value()

        # relative or absolute
        estimate_mode = ["relative", "absolute"][
            self.estimate_drift_mode.currentIndex()
        ]

        # channel
        key_channel = self.key_channel.value()

        # get increment
        increment = self.increment_box.value()

        # get increment
        upsample_factor = self.upsample_box.value()

        use_roi = self.roi_checkbox.isChecked()

        roi = None
        if use_roi:
            roi_layer = self.viewer.layers[self.ROI_LAYER_NAME]

            if len(roi_layer.data) == 0:
                raise RuntimeWarning(
                    f"'Use ROI' is checked, but there is no shape layer with name '{self.ROI_LAYER_NAME}'"
                )

            if len(roi_layer.data) > 1:
                notifications.show_info(
                    f"More than one ROI in '{self.ROI_LAYER_NAME}' shape layer, using first..."
                )

            shape_poly = roi_layer.data[0]
            roi = ROIRect.from_shape_poly(
                shape_poly,
                self.CorrDrift.dims,
                z_min=self.roi_z_min.value(),
                z_max=self.roi_z_max.value(),
            )

        drift_shifts = self.CorrDrift.estimate_drift(
            t0=key_frame,
            channel=key_channel,
            increment=increment,
            upsample_factor=upsample_factor,
            roi=roi,
            mode=estimate_mode,
        )

        if increment > 1:
            drift_shifts = self.CorrDrift.interpolate_drift(drift_shifts)

        self.drift_table = TableWidget(self.viewer, self.CorrDrift)
        self.drift_table.set_content(drift_shifts)

        self.viewer.window.add_dock_widget(
            self.drift_table,
            area="right",
            name="Estimated Drift",
            tabify=False,
        )

    def load_drift(self, filename=None):
        if filename is not None:
            filename, _ = QFileDialog.getOpenFileName(
                self, "Save as csv...", ".", "*.csv"
            )

        self._check_input()

        self.drift_table = TableWidget(self.viewer, self.CorrDrift)
        self.drift_table.load(filename=filename)

        self.viewer.window.add_dock_widget(
            self.drift_table,
            area="right",
            name=f"Drift ({filename})",
            tabify=False,
        )

    def apply_drift(self):
        if self.CorrDrift is None:
            notifications.show_warning("Estimate or load drift first")
            return

        image_corrected = self.CorrDrift.apply_drifts(
            self.drift_table.get_content(),
            extend_output=self.extend_output.isChecked(),
        )

        img_layer = self.viewer.add_image(
            image_corrected,
            name=f"{self.get_current_input_layer().name} (corrected)",
        )
        img_layer.editable = False


class TableWidget(QWidget):
    """ """

    def __init__(
        self, ist: CorrectDrift = None, viewer: "napari.Viewer" = None
    ):
        super().__init__()

        self.ist = ist
        self._napari_viewer = viewer

        self._view = QTableView()

        copy_button = QPushButton("Copy to clipboard")
        copy_button.clicked.connect(self.copy_clicked)

        save_button = QPushButton("Save as csv...")
        save_button.clicked.connect(self.save)

        self.setWindowTitle("Drift")
        self.setLayout(QGridLayout())

        action_widget = QWidget()
        action_widget.setLayout(QHBoxLayout())
        action_widget.layout().addWidget(QLabel("Drifts to apply:"))
        action_widget.layout().addWidget(copy_button)
        action_widget.layout().addWidget(save_button)
        self.layout().addWidget(action_widget)
        self.layout().addWidget(self._view)
        # action_widget.layout().setSpacing(3)
        # action_widget.layout().setContentsMargins(0, 0, 0, 0)

    def save(self, event=None, filename=None):
        if filename is None:
            filename, _ = QFileDialog.getSaveFileName(
                self, "Save as csv...", ".", "*.csv"
            )
        DataFrame(self.get_content()).to_csv(filename, header=["Z", "Y", "X"])

    def copy_clicked(self):
        DataFrame(self.get_content()).to_clipboard()

    def load(self, event=None, filename=None):
        drift_shifts = read_csv(filename, index_col=0).to_numpy()
        self.set_content(drift_shifts)

    def set_content(self, table: np.array):
        self._table = ShiftTableModel(table)
        self._view.setModel(self._table)
        self._view.resizeColumnsToContents()
        self._view.setCornerButtonEnabled(False)

    def get_content(self) -> np.array:
        return self._table._data


class ShiftTableModel(QAbstractTableModel):
    def __init__(self, data):
        super().__init__()
        self._data = data

    def rowCount(self, index):
        return self._data.shape[0]

    def columnCount(self, index):
        return self._data.shape[1]

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            return "ZYX"[section]

        if orientation == Qt.Vertical and role == Qt.DisplayRole:
            return str(section)

        return super().headerData(section, orientation, role)

    def data(self, index, role=Qt.DisplayRole):
        if index.isValid():
            if role == Qt.DisplayRole or role == Qt.EditRole:
                value = self._data[index.row(), index.column()]
                return str(value)

    def setData(self, index, value, role):
        if role == Qt.EditRole:
            try:
                value = float(value)
            except ValueError:
                return False
            self._data[index.row(), index.column()] = value
            return True
        return False

    def flags(self, index):
        return Qt.ItemIsSelectable | Qt.ItemIsEnabled | Qt.ItemIsEditable



import warnings

import numpy as np
from napari.utils import progress
from scipy.interpolate import interp1d
from scipy.ndimage import shift
from skimage.registration import phase_cross_correlation


class ArrayAxesStandardizer:
    """
    A class designed to standardize the axes of numpy arrays.

    Attributes:
        out_order (str): string representing the desired output
        order of the array axes in_order (str): string representing
        the current order of the array axes

        out_order and in_order are given as string containing
        t,c,z,y,x

    Methods:
        __init__(self, out_order: str, in_order: str):
            Initializes the class and checks for any invalid inputs.

        _check_order_str(self, order: str):
            Checks for any duplicates in the input order.

        __call__(self, data: np.array):
            Takes in an input numpy array and standardizes the
            axes according to the output order.
            It returns the standardized array view.

        inv(self, data: np.array):
            Takes in a standardized numpy array and reverts the
            axes back to the original order.
            It returns the reverted array.

        inverse_permutation(a):
            A static method that returns the inverse permutation
            of an input permutation.

    Parameters:
        out_order (str): a string representing the desired output
        order of the array axes.
        in_order (str): a string representing the current order
        of the array axes.
        data (np.array): a numpy array that needs to be
        standardized or reverted.
        a (np.array): an input permutation.

    Return Value:
        __call__(self, data: np.array): A standardized numpy array.
        inv(self, data: np.array): A numpy array with the original
        axis order.
        inverse_permutation(a): An inverse permutation of the input
        permutation.
    """

    def __init__(self, out_order: str, in_order: str):
        """
        Initializes the class and checks for any invalid inputs.

        Args:
            out_order (str): A string representing the desired output
            order of the array axes.
            in_order (str): A string representing the current order of
            the array axes.

        Raises:
            AssertionError: If in_order contains any elements not in
            out_order or if out_order or in_order contains duplicates.
        """
        self._check_order_str(out_order)
        self._check_order_str(in_order)

        assert set(in_order).issubset(
            set(out_order)
        ), "in_order not subset of out_order"

        self.out_order = out_order
        self.in_order = in_order

    def _check_order_str(self, order: str):
        """
        Checks for any duplicates in the input order.

        Args:
            order (str): A string representing the current order of
            the array axes.

        Raises:
            AssertionError: If order contains duplicates.
        """
        assert len(set(order)) == len(
            order
        ), f"Duplicates in order found: '{order}'"

    def __call__(self, data: np.array):
        """
        Takes in an input numpy array and standardizes the axes according
        to the output order.
        It returns the standardized array.

        Args:
            data (np.array): A numpy array that needs to be standardized.

        Returns:
            np.array: A standardized numpy array.

        Raises:
            AssertionError: If the shape of the input array does not match
            the input order.
        """

        assert len(data.shape) == len(self.in_order), (
            f"Shape mismatch. Data shape is '{data.shape}'"
            ", but in order is '{self.in_order}'"
        )
        permute = []
        missing = []
        for i, dim in enumerate(self.out_order):
            j = self.in_order.find(dim)
            permute.append(j) if j > -1 else missing.append(i)

        data_rearranged = np.expand_dims(data.transpose(permute), missing)
        # assert not data_rearranged.flags["OWNDATA"], "not a view??"
        return data_rearranged

    def inv(self, data: np.array):
        """
        Applies the inverse transform to the input data. The input data
        is expected to be a numpy array with the shape
        specified by the 'out_order' argument passed to the constructor.
        The output is a numpy array with the shape
        specified by the 'in_order' argument passed to the constructor.

        Args:
            data (np.array): The input data to be transformed.

        Returns:
            np.array: The transformed data.

        Raises:
            AssertionError: If the input data does not have the same shape
            as specified by the 'out_order' argument.
        """

        assert len(data.shape) == len(self.out_order), (
            f"Shape mismaatch. Data shape is '{data.shape}',"
            " but in order is '{self.out_order}'"
        )
        permute = []
        missing = []
        for i, dim in enumerate(self.out_order):
            j = self.in_order.find(dim)
            permute.append(j) if j > -1 else missing.append(i)

        res = np.transpose(
            np.squeeze(data, axis=tuple(missing)),
            self.inverse_permutation(permute),
        )
        return res

    @staticmethod
    def inverse_permutation(a):
        b = np.arange(len(a))
        b[a] = b.copy()
        return b


class ROIRect:
    def __init__(
        self,
        x_min: int,
        x_max: int,
        y_min: int,
        y_max: int,
        z_min: int,
        z_max: int,
        t0: int,
        c0: int,
    ):
        self.x_min = x_min
        self.x_max = x_max

        self.y_min = y_min
        self.y_max = y_max

        self.z_min = z_min
        self.z_max = z_max

        self.t0 = t0
        self.c0 = c0

        self._check()

    def _check(self):
        assert (
            self.z_max > self.z_min > -1
        ), "Z-dim mismatch. Choose Z minimum and maximum"
        assert self.y_max > self.y_min > -1, "y-dim mismatch"
        assert self.x_max > self.x_min > -1, "x-dim mismatch"

    @classmethod
    def from_shape_poly(
        cls, shape_poly: np.array, dims: str, z_min: int, z_max: int
    ):
        p_min = shape_poly.min(0)
        p_max = shape_poly.max(0)

        t = dims.find("t")
        t0 = int(p_min[t])

        x_min = int(p_min[-1] + 0.5)
        x_max = int(p_max[-1] + 0.5)

        y_min = int(p_min[-2] + 0.5)
        y_max = int(p_max[-2] + 0.5)

        if "z" not in dims:
            z_min = 0
            z_max = 1

        c0 = 0
        if "c" in dims:
            c = dims.find("c")
            c0 = int(p_min[c])

        return cls(x_min, x_max, y_min, y_max, z_min, z_max, t0, c0)

    @classmethod
    def from_bbox(cls, bbox: np.array, t0: int, c0: int):
        return cls(
            bbox[4], bbox[5], bbox[2], bbox[3], bbox[0], bbox[1], t0, c0
        )

    @property
    def origin(self):
        return self.bbox[::2]

    @property
    def shape(self):
        return self.bbox[1::2] - self.bbox[::2]

    @property
    def bbox(self):
        return np.array(
            [
                self.z_min,
                self.z_max,
                self.y_min,
                self.y_max,
                self.x_min,
                self.x_max,
            ]
        )

    def __str__(self):
        return f"""
        t0 {self.t0}
        c0 {self.c0}
        z: {self.z_min}, {self.z_max}
        y: {self.y_min}, {self.y_max}
        x: {self.x_min}, {self.x_max}

        bbox: {self.bbox}
        """


class CorrectDrift:
    """
    Class for drift correction of microscopy images.

    Parameters
    ----------
    data : np.array
        The input data, with dimensions ordered according to
        the given `dims`.
    dims : str
        The dimension order of the input data. Must include
        axes 't', 'x', and 'y'.

    Attributes
    ----------
    is_multi_channel : bool
        True if the data has multiple channels.
    is_3d : bool
        True if the data has a z dimension.
    dims : str
        The dimension order of the input data.
    data_arranger : ArrayAxesStandardizer
        An ArrayAxesStandardizer instance used to rearrange the input data.
    data : np.array
        The rearranged input data.
    T, C, Z, Y, X : int
        The number of time points, channels, z-slices, rows, and columns
        in the input data.

    Methods
    -------
    estimate_shifts_absolute(t0=0, channel=0, increment=1, upsample_factor=1,
    roi=None)
        Estimates the absolute shifts for each time point using cross-
        correlation.
    iter_abs(T, t0, step)
        An iterator that generates time points for estimating absolute shifts.
    estimate_shifts_relative(t0=0, step=1, upsample_factor=1, roi=None)
        Estimates the relative shifts for each time point using cross-
        correlation.
    iter_rel(T, t0, step)
        An iterator that generates time points for estimating relative shifts.
    interpolate_offsets(offsets)
        Interpolates any missing offsets using linear interpolation.
    apply_shifts(offsets, extend_output=False, order=1, mode='constant')
        Applies the given offsets to the input data to correct for drift.

    """

    def __init__(self, data: np.array, dims: str):
        assert (
            "t" in dims
        ), f"Axis 't' for stabilizing not found in data dims: '{dims}'"
        assert (
            "x" in dims and "y" in dims
        ), f"Axis x or y not present in dims: '{dims}'"

        self.is_multi_channel = "c" in dims
        self.is_3d = "z" in dims

        self.dims = dims
        self.data_arranger = ArrayAxesStandardizer("tczyx", dims)
        self.data = self.data_arranger(data)

        self.T, self.C, self.Z, self.Y, self.X = self.data.shape

    @staticmethod
    def iter_abs(T: int, t0: int, step: int) -> np.array:
        rm = np.c_[np.ones(T) * t0, np.arange(T)]
        rm_inc = rm[::step, :]

        # make sure last element is in iter (needed for interpolation)
        if not np.all(rm[-1] == rm_inc[-1]):
            rm = np.r_[rm_inc, rm[-1][None]]

        return rm.astype("int32")

    @staticmethod
    def iter_rel(T: int, t0: int, step: int) -> np.array:
        rm = []

        # forward
        r = t0
        while True:
            m = r + step
            if m < T:
                rm.append([r, m])
                r += step
            else:
                break
        # backward
        r = t0
        while True:
            m = r - step
            if m > -1:
                rm.append([r, m])
                r -= step
            else:
                break
        # make sure t0 is reference, if t0 < increment
        if 1 <= t0 < step:
            rm.append([t0, 0])

        return np.array(rm)

    def estimate_drift(
        self,
        t0: int = 0,
        channel: int = 0,
        increment: int = 1,
        upsample_factor: int = 1,
        roi: ROIRect = None,
        mode: str = "relative",
    ):
        if mode == "relative":
            return self._estimate_drift_relative(
                t0, channel, increment, upsample_factor, roi
            )

        elif mode == "absolute":
            return self._estimate_drift_absolute(
                t0, channel, increment, upsample_factor, roi
            )

        else:
            raise AttributeError(f"Estimation mode '{mode}' not valid")

    def _estimate_drift_relative(
        self,
        t0: int = 0,
        channel: int = 0,
        increment: int = 1,
        upsample_factor: int = 1,
        roi: ROIRect = None,
    ):
        offsets_rel = np.zeros((self.T, 3))
        offsets_rel.fill(np.nan)

        if roi is not None:
            t0 = roi.t0
            c0 = roi.c0
            roi_t0 = ROIRect.from_bbox(roi.bbox, t0, c0)

        if not self.is_multi_channel:
            channel = 0

        for r, m in progress(
            self.iter_rel(self.T, t0, increment),
            desc=f"Estimate drift (relative to frame '{t0}')",
        ):
            if (r == t0) and (roi is not None):
                # reset ROI, now it goes downward
                mov_bbox = roi_t0.bbox.copy()

            ref_img = self.data[r, channel]
            mov_img = self.data[m, channel]

            if roi is not None:
                ref_img_crop = ref_img[
                    slice(
                        max(0, mov_bbox[0]), min(ref_img.shape[0], mov_bbox[1])
                    ),
                    slice(
                        max(0, mov_bbox[2]), min(ref_img.shape[1], mov_bbox[3])
                    ),
                    slice(
                        max(0, mov_bbox[4]), min(ref_img.shape[2], mov_bbox[5])
                    ),
                ].copy()

                ref_img = np.zeros_like(ref_img)

                ref_img[
                    : ref_img_crop.shape[0],
                    : ref_img_crop.shape[1],
                    : ref_img_crop.shape[2],
                ] = ref_img_crop

            with warnings.catch_warnings():
                warnings.filterwarnings(
                    action="ignore",
                    category=RuntimeWarning,
                )
                offset, _, _ = phase_cross_correlation(
                    ref_img,
                    mov_img,
                    upsample_factor=upsample_factor,
                    return_error="always",
                    disambiguate=True,
                )

            offset = np.asarray(offset, dtype="float32")

            if roi is not None:
                offset += mov_bbox[::2]

            if roi is not None:
                mov_bbox[::2] -= np.round(offset).astype("int32")
                mov_bbox[1::2] -= np.round(offset).astype("int32")

                mov_bbox[0:2] = np.clip(mov_bbox[0:2], 0, ref_img.shape[0])
                mov_bbox[2:4] = np.clip(mov_bbox[2:4], 0, ref_img.shape[1])
                mov_bbox[4:] = np.clip(mov_bbox[4:], 0, ref_img.shape[2])
            if m > r:
                offsets_rel[m] = -offset
            else:
                offsets_rel[r] = offset

        offsets_rel[0] = 0
        nan_row = np.nonzero(np.isnan(offsets_rel).any(axis=1))[0]
        offsets = np.nancumsum(offsets_rel, axis=0)
        offsets[nan_row, :] = np.nan

        offsets -= offsets[t0]

        return offsets

    def _estimate_drift_absolute(
        self,
        t0: int = 0,
        channel: int = 0,
        increment: int = 1,
        upsample_factor: int = 1,
        roi: ROIRect = None,
    ):
        if not self.is_multi_channel:
            channel = 0

        offsets = np.zeros((self.T, 3))
        offsets.fill(np.nan)

        ref_img = self.data[t0, channel]

        if roi is not None:
            ref_img_crop = ref_img[
                slice(max(0, roi.bbox[0]), min(ref_img.shape[0], roi.bbox[1])),
                slice(max(0, roi.bbox[2]), min(ref_img.shape[1], roi.bbox[3])),
                slice(max(0, roi.bbox[4]), min(ref_img.shape[2], roi.bbox[5])),
            ].copy()

            ref_img = np.zeros_like(ref_img)

            ref_img[
                : ref_img_crop.shape[0],
                : ref_img_crop.shape[1],
                : ref_img_crop.shape[2],
            ] = ref_img_crop

        for _, m in progress(
            self.iter_abs(self.T, t0, increment),
            desc=f"Estimate drift (absolute to frame '{t0}')",
        ):
            mov_img = self.data[m, channel]

            with warnings.catch_warnings():
                warnings.filterwarnings(
                    action="ignore",
                    category=RuntimeWarning,
                )
                offset, _, _ = phase_cross_correlation(
                    ref_img,
                    mov_img,
                    upsample_factor=upsample_factor,
                    return_error="always",
                    disambiguate=True,
                )

            offset = -np.asarray(offset, dtype="float32")

            if roi is not None:
                offset -= roi.bbox[::2]

            offsets[m] = offset

        return offsets

    def interpolate_drift(self, offsets: np.array):
        x = np.nonzero(~np.isnan(offsets).any(axis=1))[0]
        m = np.nonzero(np.isnan(offsets).any(axis=1))[0]
        y = offsets[x, :]
        offsets[m] = interp1d(
            x, y, kind="linear", axis=0, fill_value="extrapolate"
        )(m)

        return offsets

    def apply_drifts(
        self,
        offsets: np.array,
        extend_output: bool = False,
        order: int = 1,
        mode: str = "constant",
    ):
        if extend_output:
            # compute new shape of extended output
            shape_ext_zyx = np.ceil(offsets.max(0) - offsets.min(0)).astype(
                int
            )
            shape_yxz = np.array([self.Z, self.Y, self.X])
            out_shape_zyx = shape_ext_zyx + shape_yxz

            # create extended output
            output = np.zeros(
                (self.T, self.C) + tuple(out_shape_zyx), dtype="float32"
            )

            # split shifts into integer and subpixel parts.
            # the ndi.shift takes care of subpixel part
            # the interger shifts are handled via giving an output slice

            offsets_px = -np.ceil(offsets - offsets.max(0)).astype("int")
            offsets_sub = -(offsets - offsets.max(0)) - offsets_px

            for t in progress(range(self.T), desc="Applying drift"):
                for c in range(self.C):
                    img = self.data[t, c]
                    output_view = output[
                        t,
                        c,
                        offsets_px[t, 0] : offsets_px[t, 0] + img.shape[0],
                        offsets_px[t, 1] : offsets_px[t, 1] + img.shape[1],
                        offsets_px[t, 2] : offsets_px[t, 2] + img.shape[2],
                    ]

                    shift(
                        img,
                        -offsets_sub[t],
                        output=output_view,
                        order=order,
                        mode=mode,
                        prefilter=False,
                    )

        else:
            output = np.zeros_like(self.data)
            for t in progress(range(self.T), desc="Applying drift"):
                for c in range(self.C):
                    img = self.data[t, c]
                    output[t, c] = shift(
                        img,
                        -offsets[t],
                        order=order,
                        mode=mode,
                        prefilter=False,
                    )

        # Transfrom from TCZYX to original data layout
        return self.data_arranger.inv(output)
